package com.mysimpatico.sqlwrapper;

import com.mysimpatico.sqlwrapper.SqlWrapper;

/**
 * 
 * @author Gabriele Kahlout
 * Class for immutable object Table.
 *
 */

public class Table implements SelectableFrom {
	private final String name;
	private final Column[] allColumns;
	private final Column[] columns;
	private final Column[] primaryKeys;
	private final boolean foreignKeys;
	private final ReferencedColumn ridColumn = new ReferencedColumn("RID");
	private final Table refTable;
	private final String condition;
		
	public Table(final String name, final Column[] columns){
		this.name = name;
		this.columns = columns;
		this.allColumns = new Column[columns.length + 1];
		this.allColumns[columns.length] = ridColumn;
		for (int i=0; i < columns.length; i++) this.allColumns[i] = columns[i];
		
		primaryKeys = new Column[]{ridColumn};
		foreignKeys = false;
		refTable = null;
		condition = null;
	}
	
	public Table(final String name, final Column column){
		this(name, new Column[]{column});
	}
	
	public Table(final String name, final Column column, final Column primaryKey){
		this(name, new Column[]{column}, new Column[]{primaryKey});
	}
	
	public Table(final String name, final Column[] columns, final Column[] primaryKeys, 
			final boolean foreignKeys){
		this.name = name;
		this.columns = columns;
		this.allColumns = new Column[columns.length + 1];
		this.allColumns[columns.length] = ridColumn;
		for (int i=0; i < columns.length; i++) this.allColumns[i] = columns[i];
		this.primaryKeys = primaryKeys;
		this.foreignKeys = foreignKeys;
		refTable = primaryKeys[0].getRefTable();
		condition = null;
	}
	
	public Table(final String name, final Column[] columns, final Table refTable, final Column[] primaryKeys){
		this.name = name;
		this.columns = columns;
		this.allColumns = new Column[columns.length + 1];
		this.allColumns[columns.length] = ridColumn;
		for (int i=0; i < columns.length; i++) this.allColumns[i] = columns[i];
		this.primaryKeys = primaryKeys;
		foreignKeys = true;
		this.refTable = refTable;
		condition = null;

	}
	
	public Table(final String name, final Column[] columns, final Table refTable, final Column primaryKey){
		this.name = name;
		this.columns = columns;
		this.allColumns = new Column[columns.length + 1];
		this.allColumns[columns.length] = ridColumn;
		for (int i=0; i < columns.length; i++) this.allColumns[i] = columns[i];
		primaryKeys = new Column[1];
		primaryKeys[0] = primaryKey;
		foreignKeys = true;
		this.refTable = refTable;
		condition = null;

	}
	
	public Table(final String name, final Column[] columns, final Column[] primaryKeys){
		this.name = name;
		this.columns = columns;
		allColumns = new Column[columns.length + 1];
		allColumns[columns.length] = ridColumn;
		for (int i=0; i < columns.length; i++) this.allColumns[i] = columns[i];
		this.primaryKeys = primaryKeys;
		foreignKeys = false;
		refTable = null;
		condition = null;
	}
	
	public Table(final String name, final Column[] columns, final Column[] primaryKeys, final String checkQuery){
		this.name = name;
		this.columns = columns;
		allColumns = new Column[columns.length + 1];
		allColumns[columns.length] = ridColumn;
		for (int i=0; i < columns.length; i++) this.allColumns[i] = columns[i];
		this.primaryKeys = primaryKeys;
		foreignKeys = false;
		refTable = null;
		condition = checkQuery;
	}
	
	
	public Table(final String name, final Column[] columns, final Column primaryKey){
		this(name, columns, new Column[]{primaryKey});
	}
		
	public Table(final String name, final Column[] columns, final boolean exists, final String checkQuery) {
		this.name = name;
		this.columns = columns;
		allColumns = new Column[columns.length + 1];
		allColumns[columns.length] = ridColumn;
		for (int i=0; i < columns.length; i++) this.allColumns[i] = columns[i];
		condition = ((exists)? "": "NOT") + " EXISTS( " + checkQuery + ")";
		refTable = null;
		primaryKeys = null;
		foreignKeys = false;
	}
	
	public Table(final String name, final Column col, final boolean exists, String checkQuery) {
		this(name, new Column[]{col}, exists, checkQuery);
	}

	public String getName(){
		return name;
	}
	
	public int size(){
		return allColumns.length;
	}
	
	public int defsize(){
		int counter = 0;
		for (final Column c : allColumns)
			if (c.getDef() != null) counter ++;
			/*else if (c.refField != null) 
					if (((Column) c).refField.getDef() != null) 
						counter++;
		}*/
		return counter;
	}
	
	private String getForeign(){
		final String foreign;
		return (foreignKeys)? ", FOREIGN KEY (" + (foreign = SqlWrapper.joinColumns(primaryKeys,SqlWrapper.QueryType.FOREIGN)) + ") REFERENCES " +  refTable.getName() + " ("+foreign+ ")" : "";
	}
	
	private String getPrimaryKey(){
		return (primaryKeys == null)? "" :" " + ((SqlWrapper.db.equals(SqlWrapper.vendor.SQLITE))? ", UNIQUE ("  + SqlWrapper.joinColumns(primaryKeys,SqlWrapper.QueryType.FOREIGN)+ ") CHECK("+SqlWrapper.joinColumns(primaryKeys, SqlWrapper.QueryType.PRIMARY) + ")" 
				: ", PRIMARY KEY("  + SqlWrapper.joinColumns(primaryKeys,SqlWrapper.QueryType.FOREIGN)+ ")");
	}
	
	private String getCondition(){
		return (condition == null)? "" : ", CHECK( " + condition + ")";
	}
	
	@Override
	public final String toString(){
		return name + " (" + SqlWrapper.joinColumns(allColumns,SqlWrapper.QueryType.CREATE) + getPrimaryKey() + getForeign() + getCondition() + ")";
		
		/* if (foreignKeys) return name + " (" + SqlWrapper.joinColumns(allColumns,SqlWrapper.QueryType.CREATE) + ", FOREIGN KEY (" + (foreign = SqlWrapper.joinColumns(primaryKeys,SqlWrapper.QueryType.FOREIGN)) + ") REFERENCES " +  refTable.getName() + " ("+foreign+ "))";
		return (primaryKeys == null)? name + " (" + SqlWrapper.joinColumns(allColumns,SqlWrapper.QueryType.CREATE) + ")":
			name + " (" + SqlWrapper.joinColumns(allColumns,SqlWrapper.QueryType.CREATE) + ", UNIQUE ("  + SqlWrapper.joinColumns(primaryKeys,SqlWrapper.QueryType.FOREIGN)+ ") CHECK("+SqlWrapper.joinColumns(primaryKeys, QueryType.PRIMARY) + "))"; */ 
	}
	
	public Column[] getColumns(){
		return columns;
	}
	
	public Column[] getAllColumns(){
		return allColumns;
	}

	public ReferencedColumn getIdColumn() {
		return ridColumn;
	}

	@Override
	public boolean supportBoolean() {
		for (Column c: allColumns) if (c.getType().equals(SqlWrapper.Type.BOOLEAN)) return true;
		return false;
	}
}
